open Okweb
open Core
open Lwt.Infix

module Impl = struct
  module type S = sig
    type auth_storage
    val process_request: auth_storage Okauth_client.config -> Handler.t
  end

  module Make
      (Entry_repo: Data.Entry_repo.S)
      (St: Okauth_client.Storage)
      (Auth: Okauth_client.S with type storage := St.t) = struct
    let get_months_of year dates_with_entries =
      let calendar = List.map Month.all ~f:begin fun month ->
          let days_max = Date.days_in_month ~year ~month in
          let shift =
            Date.create_exn ~y:year ~m:month ~d:1
            |> Date.day_of_week
            |> Day_of_week.iso_8601_weekday_number
            |> Int.pred in
          let weeks_num =
            (Float.of_int (days_max + shift) /. 7.0) |> Float.iround_up_exn in
          let weeks = List.range 0 weeks_num |> List.map ~f:begin fun week_n ->
              let days = List.range 0 7 |> List.map ~f:begin fun day_n ->
                  match (week_n * 7) + day_n - shift + 1 with
                  | day_of_month when day_of_month > 0 && day_of_month <= days_max ->
                    let full_date = Date.create_exn ~y:year ~m:month ~d:day_of_month in
                    `O [ "text", `String (Int.to_string day_of_month);
                         "date", `String (Printf.sprintf "%04d%02d%02d" year (Month.to_int month) day_of_month);
                         "has_entry", `Bool (Base.Hash_set.mem dates_with_entries full_date) ]
                  | _ ->
                    `Null
                end in
              `O [ "days", `A days ]
            end in
          `O [ "name", `String (Month.to_string month);
               "weeks", `A weeks ]
        end in
      `A calendar

    let get_calendar user_id year =
      let dates_with_entries = Entry_repo.find_dates_with_entries user_id year in
      let template = Mustache.of_string [%blob "web/calendar_page.html"] in
      let params = `O [
          "prev_year", `String (Int.(to_string (pred year)));
          "this_year", `String (Int.to_string year);
          "next_year", `String (Int.(to_string (succ year)));
          "months", get_months_of year dates_with_entries
        ] in
      let body = Mustache.render template params in
      Response.(
        ok
        |> with_header ~key:"Content-Type" ~value:"text/html;charset=utf-8"
        |> with_string_body body)

    let get_entry user_id date =
      let template = Mustache.of_string [%blob "web/entry_page.html"] in
      let params = match Entry_repo.get_by_user_id_and_date user_id date with
        | Some entry ->
          `O [ "date", `String (Date.format date "%b %e, %Y");
               "content", `String entry.Data.Entry.content ]
        | None ->
          `O [ "date", `String (Date.format date "%b %e, %Y");
               "content", `String "" ]
      in
      let body = Mustache.render template params in
      Response.(
        ok
        |> with_header ~key:"Content-Type" ~value:"text/html;charset=utf-8"
        |> with_string_body body)

    let save_entry user_id date content =
      ignore (Entry_repo.create_or_update ~user_id ~date ~content);
      Response.redirect (Printf.sprintf "/entry/%s" (Date.to_string_iso8601_basic date))

    let do_process_request req =
      let user_id_opt = Request.get_cookies req |> Cookie_jar.find_value ~key:"user_id" in
      match Request.(get_http_method req, get_path req, user_id_opt) with
      | "GET", ["okdiary.css"], _ ->
        Lwt.return Response.(
            ok
            |> with_header ~key:"Content-Type" ~value:"text/css"
            |> with_string_body [%blob "web/okdiary.css"])
      | "GET", ["okdiary.js"], _ ->
        Lwt.return Response.(
            ok
            |> with_header ~key:"Content-Type" ~value:"text/javascipt"
            |> with_string_body [%blob "web/okdiary.js"])
      | "GET", [], _ ->
        Lwt.return (Response.redirect "/calendar")
      | "GET", ["calendar"], Some user_id ->
        let current_year = Date.today ~zone:Time.Zone.utc |> Date.year in
        Lwt.return (get_calendar user_id current_year)
      | "GET", ["calendar"; year_str], Some user_id ->
        Lwt.return begin match int_of_string_opt year_str with
          | Some year when year > 1900 ->
            get_calendar user_id year
          | _ ->
            Response.(of_status 400 |> with_string_body "Invalid year")
        end
      | "GET", ["entry"; date_str], Some user_id ->
        let date_opt = Option.try_with (fun () -> Date.of_string_iso8601_basic date_str ~pos:0) in
        Lwt.return begin match date_opt with
          | Some date -> get_entry user_id date
          | _ -> Response.(of_status 400 |> with_string_body "Invalid date")
        end
      | "POST", ["entry"; date_str], Some user_id ->
        let date_opt = Option.try_with (fun () -> Date.of_string_iso8601_basic date_str ~pos:0) in
        Request.get_body req
        |> Form.of_body
        >|= Form.find_param ~key:"content"
        >|= fun content_opt -> begin match date_opt, content_opt with
          | Some date, Some content -> save_entry user_id date content
          | _ -> Response.(of_status 400 |> with_string_body "Invalid request")
        end
      | _ ->
        Lwt.return Response.(of_status 404 |> with_string_body "Not found")

    let process_request auth_config =
      let auth_middleware =
          if Option.equal String.equal
            (Sys.getenv "OKDIARYOPTS")
            (Some "i_know_what_i_am_doing_disable_authentication_please")
          then Fn.id
          else Auth.middleware auth_config
      in
      do_process_request
      |> auth_middleware
      |> Looksok.middleware
  end
end


module type S = sig
  module type S = sig
    include Impl.S
  end

  [@@@warning "-67"]
  module Make
      (Entry_repo: Data.Entry_repo.S)
      (St: Okauth_client.Storage)
      (Auth: Okauth_client.S with type storage := St.t):
    S with type auth_storage := St.t
end
